from datetime import datetime
from dateutil.relativedelta import relativedelta
from throttleandcache.tdparse import parse


def test_seconds():
    assert parse('1s') == relativedelta(seconds=1)
    assert parse('1 second') == relativedelta(seconds=1)
    assert parse('2 seconds') == relativedelta(seconds=2)


def test_minutes():
    assert parse('1m') == relativedelta(minutes=1)
    assert parse('1min') == relativedelta(minutes=1)
    assert parse('1 minute') == relativedelta(minutes=1)
    assert parse('2 minutes') == relativedelta(minutes=2)


def test_hours():
    assert parse('1h') == relativedelta(hours=1)
    assert parse('1hr') == relativedelta(hours=1)
    assert parse('1 hour') == relativedelta(hours=1)
    assert parse('2 hours') == relativedelta(hours=2)


def test_mixed():
    assert parse('3 minutes, 2 seconds') == relativedelta(minutes=3, seconds=2)


def test_number():
    """
    Seconds should be assumed when a number is passed.

    """
    assert parse(1) == relativedelta(seconds=1)


def test_offset():
    # Make sure the parsed value can be added to a datetime. This is important
    # because it's possible to create valid relativedeltas that can't be added
    # to datetimes. (e.g. with a float value for "year")
    datetime.now() + parse('1s')



from dateutil.relativedelta import relativedelta
import re


class ParseError(Exception):
    pass


PATTERN = r"""
    (.*?(?P<years>\d+)\s*y)?
    (.*?(?P<months>\d+)\s*(mon|mos))?
    (.*?(?P<weeks>\d+(\.\d+)?)\s*w)?
    (.*?(?P<days>\d+(\.\d+)?)\s*d)?
    (.*?(?P<hours>\d+(\.\d+)?)\s*h)?
    (.*?(?P<minutes>\d+(\.\d+)?)\s*m)?
    (.*?(?P<seconds>\d+(\.\d+)?)\s*s)?
"""

COMPILED_PATTERN = re.compile(PATTERN, re.VERBOSE)


def parse(val):
    try:
        return relativedelta(seconds=float(val))
    except ValueError:
        pass

    match = COMPILED_PATTERN.search(val)
    if not match:
        raise ParseError("Couldn't parse %s" % val)
    kwargs = dict((k, float(v or 0)) for k, v in match.groupdict().items())
    for attr in ['years', 'months']:
        # Years and months don't work with floats. (You can create the
        # relativedelta, but it can't be successfully added to a datetime.)
        kwargs[attr] = int(kwargs[attr])
    return relativedelta(**kwargs)



from django.core.cache import cache as cache_obj
from mock import Mock
from throttleandcache import cache


SOME_VALUE = {'hello': 'world'}


def teardown_function(fn):
    cache_obj.clear()


def test_none():
    fn = cache()(lambda: None)
    assert fn() is None


def test_caching():
    """
    Makes sure the cache is actually used.
    """
    f = Mock(__name__='f', return_value=None)
    decorated = cache()(f)
    decorated()
    decorated()
    assert f.call_count == 1


def test_value():
    """
    Makes sure the cached value is correct.
    """
    f = Mock(__name__='f', return_value=SOME_VALUE)
    decorated = cache()(f)
    assert decorated() == SOME_VALUE


def test_nocache():
    """
    Makes sure that the cache decorator doesn't cache when it has a 0
    timeout.
    """
    f = Mock(__name__='f', return_value=None)
    decorated = cache(timeout=0)(f)
    decorated()
    decorated()
    assert f.call_count == 2


def test_graceful():
    def f():
        if f.called:
            f.errored = True
            raise Exception
        f.called = True
        return SOME_VALUE

    f.called = False
    f.errored = False
    decorated = cache('0s', graceful=True)(f)
    decorated()  # Successful call
    value = decorated()  # Unsuccessful, but should be handled gracefully.
    assert f.errored
    assert value == SOME_VALUE



from appconf import AppConf
from django.conf import settings


class ThrottleAndCacheConf(AppConf):
    DEFAULT_CACHE = None
    MAX_TIMEOUT = 60 * 60 * 24 * 365 * 10

    def configure_default_cache(self, value):
        if value is None:
            try:
                from django.core.cache.backends.dummy import DummyCache
            except ImportError:
                dummy_cache = 'dummy://'
            else:
                dummy_cache = 'django.core.cache.backends.dummy.DummyCache'

            # DEFAULT_CACHE_ALIAS doesn't exist in Django<=1.2
            try:
                from django.core.cache import DEFAULT_CACHE_ALIAS as default_cache_alias
            except ImportError:
                default_cache_alias = 'default'

            if settings.DEBUG:
                value = dummy_cache
            elif default_cache_alias in getattr(settings, 'CACHES', {}):
                value = default_cache_alias
            else:
                value = getattr(settings, 'CACHE_BACKEND', None) or dummy_cache

        return value



from __future__ import absolute_import
import logging


try:
    from logging import NullHandler
except ImportError:
    from logging import Handler

    class NullHandler(Handler):
        def emit(self, record):
            pass


logger = logging.getLogger('throttleandcache')
logger.addHandler(NullHandler())



SECRET_KEY = '2@4p=l!t7c1d@pmo%aavvm_ngzyi8=$!lmfhkk@alxc=0q(##u'

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'test-cache',
    }
}



import os
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys


read = lambda fname: open(os.path.join(os.path.dirname(__file__), fname)).read()


README = read('README.markdown')


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['tests', '-s']
        self.test_suite = True

    def run_tests(self):
        import pytest
        os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.settings'
        errno = pytest.main(self.test_args)
        sys.exit(errno)


# Load package meta from the pkgmeta module without loading throttleandcache.
pkgmeta = {}
execfile(os.path.join(os.path.dirname(__file__),
         'throttleandcache', 'pkgmeta.py'), pkgmeta)


setup(
    name='django-throttleandcache',
    version=pkgmeta['__version__'],
    description='A utility for caching/throttling function calls.',
    url='http://github.com/matthewwithanm/django-throttleandcache',
    license='BSD',
    long_description=README,
    author=pkgmeta['__author__'],
    author_email='m@tthewwithanm.com',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'python-dateutil<2.0a',
        'django-appconf>=0.5',
    ],
    tests_require=[
        'pytest==2.3.5',
        'mock==1.0.1',
        'Django',
    ],
    cmdclass={'test': PyTest},
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



from hashlib import sha256
from datetime import datetime
from django.core.cache import get_cache
from django.utils.decorators import method_decorator
from django.conf import settings
from functools import wraps
from time import mktime
from .logging import logger
from .tdparse import parse


def __cache_key(fn, fn_args, fn_kwargs):
    key = fn.__module__ + fn.__name__ + repr(fn_args) + repr(fn_kwargs)
    return sha256(key).hexdigest()


class CachedValue(object):
    """
    A wrapper for cached values. This allows us to store ``None``, as well as
    letting us store metadata along with the result in the cache.

    """
    def __init__(self, value, set_time, expiration_time):
        self.value = value
        self.set_time = set_time
        self.expiration_time = expiration_time


def cache(timeout=-1, using=None, key_prefix='', graceful=False):
    """
    Cache the result of a function call for <timeout> seconds.
    """

    cache_backend = get_cache(using or settings.THROTTLEANDCACHE_DEFAULT_CACHE)

    if timeout == -1:
        timeout = settings.THROTTLEANDCACHE_MAX_TIMEOUT

    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            key = key_prefix + __cache_key(fn, args, kwargs)
            cached = cache_backend.get(key)
            now = datetime.now()

            if cached is None or cached.expiration_time < now:
                # The cached value is expired (or the result was never cached)
                try:
                    result = fn(*args, **kwargs)
                except Exception as exc:
                    if graceful and cached:
                        # There was an error executing the function, but we have
                        # a cached value to fall back to. Log the error and
                        # return the cached value.

                        logger.exception(exc)
                        return cached.value
                    raise

                then = now + parse(timeout)
                if graceful:
                    # With the graceful option, we actually want to keep the
                    # result in the cache until we explicitly override it, in
                    # case we need it later. The expiration_time will be used
                    # to determine whether the value should be recalculated
                    # instead of its absence in the cache.
                    secs = settings.THROTTLEANDCACHE_MAX_TIMEOUT
                else:
                    secs = int(mktime(then.timetuple()) - mktime(now.timetuple()))
                val = CachedValue(result, now, then)
                cache_backend.set(key, val, secs)
            else:
                result = cached.value

            return result
        return wrapper
    return decorator


# Aliases for clarity
cacheforinstance = cache


def cacheforclass(*args, **kwargs):
    decorator = cache(*args, **kwargs)
    return method_decorator(decorator)


# For backwards compatibility
def cache_result(timeout=-1, cache=None, key_prefix=''):
    return cache(timeout, cache, key_prefix)
